# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.api import Client
from croud.config import get_output_format
from croud.printer import print_response


def _org_id_transform(field):
    return field or ""


def cloud_configurations_set(args: Namespace) -> None:
    body = {"value": args.value}
    if args.org_id:
        body["organization_id"] = args.org_id

    client = Client.from_args(args)
    data, errors = client.put(f"/api/v2/configurations/{args.key}/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["key", "value", "organization_id"],
        success_message="Configuration updated.",
        output_fmt=get_output_format(args),
        transforms={"organization_id": _org_id_transform},
    )


def cloud_configurations_get(args: Namespace) -> None:
    client = Client.from_args(args)
    params = {}
    if args.org_id:
        params["organization_id"] = args.org_id
    data, errors = client.get(f"/api/v2/configurations/{args.key}/", params=params)
    print_response(
        data=data,
        errors=errors,
        keys=["key", "value", "organization_id"],
        output_fmt=get_output_format(args),
        transforms={"organization_id": _org_id_transform},
    )


def cloud_configurations_list(args: Namespace) -> None:
    client = Client.from_args(args)
    params = {}
    if args.org_id:
        params["organization_id"] = args.org_id
    data, errors = client.get("/api/v2/configurations/", params=params)
    print_response(
        data=data,
        errors=errors,
        keys=["key", "value", "organization_id"],
        output_fmt=get_output_format(args),
        transforms={"organization_id": _org_id_transform},
    )



# PYTHON_ARGCOMPLETE_OK
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import sys
from distutils.util import strtobool

import colorama
import shtab

from croud.apikeys.commands import (
    api_keys_create,
    api_keys_delete,
    api_keys_edit,
    api_keys_list,
)
from croud.cloud_configurations.commands import (
    cloud_configurations_get,
    cloud_configurations_list,
    cloud_configurations_set,
)
from croud.clusters.commands import (
    clusters_delete,
    clusters_deploy,
    clusters_expand_storage,
    clusters_get,
    clusters_list,
    clusters_restart_node,
    clusters_scale,
    clusters_set_backup_schedule,
    clusters_set_deletion_protection,
    clusters_set_ip_whitelist,
    clusters_set_product,
    clusters_set_suspended,
    clusters_snapshots_list,
    clusters_snapshots_restore,
    clusters_upgrade,
    create_scheduled_job,
    delete_scheduled_job,
    edit_scheduled_job,
    export_jobs_create,
    export_jobs_delete,
    export_jobs_list,
    get_scheduled_job_log,
    get_scheduled_jobs,
    import_job_progress,
    import_jobs_create_from_azure_blob_storage,
    import_jobs_create_from_file,
    import_jobs_create_from_s3,
    import_jobs_create_from_url,
    import_jobs_delete,
    import_jobs_list,
)
from croud.config import CONFIG
from croud.config.commands import (
    config_add_profile,
    config_current_profile,
    config_remove_profile,
    config_set_profile,
    config_show,
)
from croud.login import login
from croud.logout import logout
from croud.me import me, me_edit
from croud.organizations.auditlogs.commands import auditlogs_list
from croud.organizations.commands import (
    org_credits_create,
    org_credits_edit,
    org_credits_expire,
    org_credits_list,
    org_files_create,
    org_files_delete,
    org_files_get,
    org_files_list,
    org_secrets_create,
    org_secrets_delete,
    org_secrets_list,
    organizations_create,
    organizations_delete,
    organizations_edit,
    organizations_get,
    organizations_list,
)
from croud.organizations.users.commands import (
    org_users_add,
    org_users_list,
    org_users_remove,
)
from croud.parser import Argument, create_parser
from croud.printer import print_error, print_info
from croud.products.commands import products_list
from croud.projects.commands import (
    project_create,
    project_delete,
    project_edit,
    projects_get,
    projects_list,
)
from croud.projects.users.commands import (
    project_users_add,
    project_users_list,
    project_users_remove,
)
from croud.regions.commands import regions_create, regions_delete, regions_list
from croud.subscriptions.commands import (
    subscription_delete,
    subscriptions_create,
    subscriptions_get,
    subscriptions_list,
)
from croud.tools.spinner import HALO
from croud.users.commands import users_delete, users_list
from croud.users.roles.commands import roles_list

# Arguments common to all import-job create commands
import_job_create_common_args = [
    Argument(
        "--cluster-id",
        type=str,
        required=True,
        help="The cluster the data will be imported into.",
    ),
    Argument(
        "--file-format",
        type=str,
        required=True,
        choices=["csv", "json", "parquet"],
        help="The format of the structured data in the file.",
    ),
    Argument(
        "--compression",
        type=str,
        required=False,
        choices=["gzip"],
        help="The compression method the file uses.",
    ),
    Argument(
        "--table",
        type=str,
        required=True,
        help="The table the data will be imported into.",
    ),
    Argument(
        "--create-table",
        type=lambda x: bool(strtobool(str(x))),  # noqa
        required=False,
        help="Whether the table should be created automatically"
        " if it does not exist. If true new columns will also be added when the data"
        " requires them.",
    ),
    Argument(
        "--transformations",
        type=str,
        required=False,
        help="The transformations to apply when fetching data. This is the SELECT "
        "statement from an SQL query that is executed on the internal DuckDB "
        "database that the data is loaded to before inserting into CrateDB. "
        "This can be used to apply arbitrary SQL functions on your data before "
        "inserting into CrateDB, i.e. `UNNEST()`, `SUM()` and similar.",
    ),
]

# fmt: off
command_tree = {
    "me": {
        "help": "Print information about the current logged in user.",
        "resolver": me,
        "commands": {
            "edit": {
                "help": "Edit your user data.",
                "extra_args": [
                    Argument(
                        "--email", type=str, required=True,
                        help="The new email address to use."
                    ),
                ],
                "resolver": me_edit,
            },
        }
    },
    "login": {
        "help": "Log in to CrateDB Cloud.",
        "resolver": login,
        "extra_args": [
            Argument(
                "--idp", type=str, required=True,
                choices=["cognito", "azuread", "github", "google"],
                help="The identity provider (IdP) for the login."
            ),
        ],
    },
    "logout": {"help": "Log out of CrateDB Cloud.", "resolver": logout},
    "config": {
        "help": "Manage croud configuration.",
        "commands": {
            "show": {
                "help": "Show the full configuration.",
                "resolver": config_show,
                "omit": {"sudo", "region", "format"},
            },
            "profiles": {
                "help": "Manage configuration profiles.",
                "commands": {
                    "current": {
                        "help": "Print the current profile.",
                        "resolver": config_current_profile,
                        "omit": {"sudo", "region"},
                    },
                    "use": {
                        "help": "Set the current profile.",
                        "resolver": config_set_profile,
                        "omit": {"sudo", "region"},
                        "extra_args": [
                            Argument(
                                "profile", type=str,
                                help="The name of the profile that should be used."
                            ),
                        ],
                    },
                    "add": {
                        "help": "Add a new profile.",
                        "resolver": config_add_profile,
                        "omit": {"sudo", "region", "format"},
                        "extra_args": [
                            Argument(
                                "profile", type=str,
                                help="The name of the profile that should be added."
                            ),
                            Argument(
                                "--endpoint", type=str, required=True,
                                help="The API endpoint for the profile."
                            ),
                            Argument(
                                "--format", type=str, required=False,
                                help="The output format for the profile."
                            ),
                            Argument(
                                "--region", type=str, required=False,
                                help="The region for the profile."
                            ),
                        ],
                    },
                    "remove": {
                        "help": "Remove an existing profile.",
                        "resolver": config_remove_profile,
                        "omit": {"sudo", "region", "format"},
                        "extra_args": [
                            Argument(
                                "profile", type=str,
                                help="The name of the profile that should be removed."
                            ),
                        ],
                    },
                },
            },
        },
    },
    "projects": {
        "help": "Manage projects.",
        "commands": {
            "create": {
                "help": "Create a project in the specified organization and region.",
                "extra_args": [
                    Argument(
                        "--name", type=str, required=True,
                        help="The project name to use.",
                    ),
                    Argument(
                        "--org-id", type=str, required=False,
                        help="The organization ID to use.",
                    ),
                    Argument(
                        "--backup-location-type",
                        choices=["s3"],
                        type=str,
                        required=False,
                        help="The type of the custom backup location to use. "
                             "Only 's3' currently supported. "
                             "CrateDB Edge regions only."
                    ),
                    Argument(
                        "--backup-location", type=str, required=False,
                        help="The location of where backups are to be stored, "
                             "i.e. name of s3 bucket for s3 locations. "
                             "CrateDB Edge regions only."
                    ),
                    Argument(
                        "--backup-location-access-key-id", type=str, required=False,
                        help="The AWS access key id for the given s3 bucket. "
                             "CrateDB Edge regions only."
                    ),
                    Argument(
                        "--backup-location-endpoint-url", type=str, required=False,
                        help="The URL to a S3 compatible endpoint. "
                             "CrateDB Edge regions only."
                    ),
                    Argument(
                        "--backup-location-secret-access-key", type=str, required=False,
                        help="The AWS secret access key for the given s3 bucket. "
                             "CrateDB Edge regions only."
                    ),
                ],
                "resolver": project_create,
            },
            "delete": {
                "help": "Delete the specified project.",
                "extra_args": [
                    Argument(
                        "-p", "--project-id", type=str, required=True,
                        help="The project ID to use.",
                    ),
                    Argument("-y", "--yes", action="store_true", default=False),
                ],
                "resolver": project_delete,
            },
            "edit": {
                "help": "Edit the specified project.",
                "extra_args": [
                    Argument(
                        "-p", "--project-id", type=str, required=True,
                        help="The project ID to use.",
                    ),
                    Argument(
                        "--name", type=str, required=False,
                        help="The new project name to use.",
                    ),
                ],
                "resolver": project_edit,
            },
            "get": {
                "help": (
                    "Get a project by its ID."
                ),
                "extra_args": [
                    Argument(
                        "id", type=str,
                        help="The ID of the project.",
                    ),
                ],
                "resolver": projects_get,
            },
            "list": {
                "help": (
                    "List all projects the current user has access to in "
                    "the specified region."
                ),
                "extra_args": [
                    Argument(
                        "--org-id", type=str, required=False,
                        help="The organization ID to use.",
                    ),
                ],
                "resolver": projects_list,
            },
            "users": {
                "help": "Manage users in projects.",
                "commands": {
                    "add": {
                        "help": "Add the selected user to a project.",
                        "extra_args": [
                            Argument(
                                "-p", "--project-id", type=str, required=True,
                                help="The project ID to use.",
                            ),
                            Argument(
                                "--user", type=str, required=True,
                                help="The user email address or ID to use.",
                            ),
                            Argument(
                                "--role", type=str, required=True,
                                help=(
                                    "The role FQN to use. Run `croud users roles list` "
                                    "for a list of available roles."
                                ),
                            ),
                        ],
                        "resolver": project_users_add,
                    },
                    "list": {
                        "help": "List all users within a project.",
                        "extra_args": [
                            Argument(
                                "-p", "--project-id", type=str, required=True,
                                help="The project ID to use.",
                            ),
                        ],
                        "resolver": project_users_list,
                    },
                    "remove": {
                        "help": "Remove the selected user from a project.",
                        "extra_args": [
                            Argument(
                                "-p", "--project-id", type=str, required=True,
                                help="The project ID to use.",
                            ),
                            Argument(
                                "--user", type=str, required=True,
                                help="The user email address or ID to use.",
                            ),
                        ],
                        "resolver": project_users_remove,
                    },
                },
            },
        },
    },
    "clusters": {
        "help": "Manage clusters.",
        "commands": {
            "get": {
                "help": (
                    "Get a cluster by its ID."
                ),
                "extra_args": [
                    Argument(
                        "id", type=str,
                        help="The ID of the cluster.",
                    ),
                ],
                "resolver": clusters_get,
            },
            "list": {
                "help": "List all clusters the current user has access to.",
                "extra_args": [
                    Argument(
                        "-p", "--project-id", type=str, required=False,
                        help="The project ID to use.",
                    ),
                    Argument(
                        "--org-id", type=str, required=False,
                        help="The organization ID to use.",
                    ),
                ],
                "resolver": clusters_list,
            },
            "deploy": {
                "help": "Deploy a new CrateDB cluster.",
                "extra_args": [
                    Argument(
                        "--product-name", type=str, required=True,
                        help="The product name to use.",
                    ),
                    Argument(
                        "--tier", type=str, required=True,
                        help="The product tier to use.",
                    ),
                    Argument(
                        "--unit", type=int, required=False,
                        help="The product scale unit to use.",
                    ),
                    Argument(
                        "-p", "--project-id", type=str,
                        required=False,
                        help="The project ID to use.",
                    ),
                    Argument(
                        "--org-id", type=str,
                        required=False,
                        help="The organization ID to use.",
                    ),
                    Argument(
                        "--cluster-name", type=str, required=True,
                        help="The CrateDB cluster name to use.",
                    ),
                    Argument(
                        "--version", type=str, required=True,
                        help="The CrateDB version to use.",
                    ),
                    Argument(
                        "--username", type=str, required=True,
                        help="The CrateDB username to use.",
                    ),
                    Argument(
                        "--password", type=str, required=True,
                        help="The CrateDB password to use. Minimum 24 characters.",
                    ),
                    Argument(
                        "--subscription-id", type=str, required=True,
                        help="The subscription to use for billing of this cluster.",
                    ),
                    Argument(
                        "--channel", type=str, default="stable", required=False,
                        choices=["stable", "testing", "nightly"],
                        help="The channel of the CrateDB version.",
                    ),
                    Argument(
                        "--cpus", type=float, required=False,
                        help="Number of CPU cores to allocate. Can be fractional. "
                             "CrateDB Edge regions only.",
                    ),
                    Argument(
                        "--disks", type=int, required=False,
                        help="Number of disks to attach. "
                             "CrateDB Edge regions only.",
                    ),
                    Argument(
                        "--disk-size-gb", type=int, required=False,
                        help="Size of disks to attach (in GiB).",
                    ),
                    Argument(
                        "--disk-type", type=str, required=False,
                        choices=["standard", "premium"],
                        help="Type of disks to use. "
                             "CrateDB Edge regions only.",
                    ),
                    Argument(
                        "--memory-size-mb", type=int, required=False,
                        help="Amount of memory to allocate (in MiB). "
                             "CrateDB Edge regions only.",
                    ),
                ],
                "resolver": clusters_deploy,
            },
            "scale": {
                "help": "Scale an existing CrateDB cluster.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--unit", type=int, required=True,
                        help="The product scale unit to use.",
                    ),
                ],
                "resolver": clusters_scale,
            },
            "upgrade": {
                "help": "Upgrade an existing CrateDB cluster to a later version.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--version", type=str, required=True,
                        help="The CrateDB version to use.",
                    ),
                ],
                "resolver": clusters_upgrade
            },
            "delete": {
                "help": "Delete the specified cluster.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument("-y", "--yes", action="store_true", default=False)
                ],
                "resolver": clusters_delete,
            },
            "restart-node": {
                "help": "Restart a node in a CrateDB cluster.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--ordinal", type=int, required=True,
                        help="The ordinal index of the node to restart.",
                    ),
                ],
                "resolver": clusters_restart_node,
            },
            "set-deletion-protection": {
                "help": "Set the deletion protection status of a CrateDB cluster.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--value", type=lambda x: bool(strtobool(str(x))),
                        required=True, help="The deletion protection status",
                    ),
                ],
                "resolver": clusters_set_deletion_protection,
            },
            "set-ip-whitelist": {
                "help": "Set IP Network Whitelist in CIDR format (comma-separated).",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--net", type=str,
                        required=True,
                        help="IP Network list in CIDR format (comma-separated).",
                    ),
                    Argument("-y", "--yes", action="store_true", default=False)
                ],
                "resolver": clusters_set_ip_whitelist,
            },
            "expand-storage": {
                "help": "Expand storage of a CrateDB cluster.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--disk-size-gb", type=int, required=True,
                        help="New size of attached disks (in GiB).",
                    )
                ],
                "resolver": clusters_expand_storage,
            },
            "set-suspended-state": {
                "help": "Suspend or resume a CrateDB cluster.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--value", type=lambda x: bool(strtobool(str(x))),
                        required=True, help="The suspended status.",
                    ),
                ],
                "resolver": clusters_set_suspended,
            },
            "set-product": {
                "help": "Change the cluster's product.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--product-name", type=str, required=True,
                        help="The new product name to use."
                    )
                ],
                "resolver": clusters_set_product,
            },
            "set-backup-schedule": {
                "help": "Change the cluster's backup schedule.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--backup-hours", type=str, required=True,
                        help="A list of hours in UTC time that represent when backups "
                             "will take place along the day. The list must contain at "
                             "least one value and can have up to 24, comma-separated."
                    )
                ],
                "resolver": clusters_set_backup_schedule,
            },
            "snapshots": {
                "help": "View and restore snapshots",
                "commands": {
                    "list": {
                        "help": "List the cluster snapshots available.",
                        "extra_args": [
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The CrateDB cluster ID to use.",
                            ),
                            Argument(
                                "--days-ago", type=int, required=False, default=2,
                                help="Number of days to look back.",
                            ),
                        ],
                        "resolver": clusters_snapshots_list,
                    },
                    "restore": {
                        "help": "Restore the specified snapshot.",
                        "extra_args": [
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The CrateDB cluster to restore "
                                     "the snapshot into.",
                            ),
                            Argument(
                                "--snapshot", type=str, required=True,
                                help="The snapshot to be restored.",
                            ),
                            Argument(
                                "--repository", type=str, required=True,
                                help="The repository that contains the snapshot to be "
                                     "restored.",
                            ),
                            Argument(
                                "--source-cluster-id", type=str, required=False,
                                help="The CrateDB cluster ID of the snapshot to be "
                                     "used belongs to. Must belong to the same "
                                     "organization as the target cluster. "
                                     "If not specified the ``--cluster-id`` CrateDB"
                                     " cluster will be used as the source.",
                            ),
                            Argument(
                                "--type", type=str, required=False,
                                choices=["all", "metadata", "tables", "sections"],
                                help="The type of data to be restored from the "
                                     "snapshot.",
                            ),
                            Argument(
                                "--tables", type=str, required=False,
                                help="The list of tables to restore, comma separated. "
                                     "Only valid together with ``--type tables``.",
                            ),
                            Argument(
                                "--sections", type=str, required=False,
                                help="The list of data sections to restore, comma "
                                     "separated. Only valid together with ``--type "
                                     "sections``. Valid sections are ``tables``, "
                                     "``views``, ``users``, ``privileges``, "
                                     "``analyzers`` or ``udfs``.",
                            ),
                        ],
                        "resolver": clusters_snapshots_restore,
                    },
                },
            },
            "import-jobs": {
                "help": "Manage data import jobs.",
                "commands": {
                    "delete": {
                        "help": "Cancels an already running data import job that has "
                                "not finished yet. "
                                "If the job has already finished it deletes it from "
                                "the job history.",
                        "extra_args": [
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The cluster the import job belongs to."
                            ),
                            Argument(
                                "--import-job-id", type=str,
                                required=True,
                                help="The ID of the Import Job."
                            ),
                        ],
                        "resolver": import_jobs_delete,
                    },
                    "list": {
                        "help": "Lists data import jobs that belong to a specific "
                                "cluster.",
                        "extra_args": [
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The cluster the import jobs belong to."
                            ),
                        ],
                        "resolver": import_jobs_list,
                    },
                    "create": {
                        "help": "Create a data import job for the specified cluster.",
                        "commands" : {
                            "from-url": {
                                "help": "Create a data import job on the specified "
                                        "cluster from a url.",
                                "extra_args": [
                                    # Type URL params
                                    Argument(
                                        "--url", type=str, required=True,
                                        help="The URL the import file will be read "
                                             "from."
                                    ),
                                ] + import_job_create_common_args,
                                "resolver": import_jobs_create_from_url,
                            },
                            "from-file": {
                                "help": "Create a data import job on the specified "
                                        "cluster from a file.",
                                "extra_args": [
                                    # Type file params
                                    Argument(
                                        "--file-id", type=str, required=False,
                                        help="The file ID that will be used for the "
                                             "import. If not specified then --file-path"
                                             " must be specified. "
                                             "Please refer to `croud organizations "
                                             "files` for more info."
                                    ),
                                    Argument(
                                        "--file-path", type=str, required=False,
                                        help="The file in your local filesystem that "
                                             "will be used. If not specified then "
                                             "--file-id must be specified. "
                                             "Please note the file will become visible "
                                             "under `croud organizations files list`."
                                    ),
                                ] + import_job_create_common_args,
                                "resolver": import_jobs_create_from_file,
                            },
                            "from-s3": {
                                "help": "Create a data import job on the specified "
                                        "cluster from an Amazon S3 compatible "
                                        "location.",
                                "extra_args": [
                                    # Type S3 params
                                    Argument(
                                        "--bucket", type=str, required=True,
                                        help="The name of the S3 bucket that contains "
                                             "the file to be imported."
                                    ),
                                    Argument(
                                        "--file-path", type=str, required=True,
                                        help="The absolute path in the S3 bucket that "
                                             "points to the file to be imported. "
                                             "Globbing (use of *) is allowed."
                                    ),
                                    Argument(
                                        "--secret-id", type=str, required=True,
                                        help="The secret that contains the access key "
                                             "and secret key needed to access the file "
                                             "to be imported."
                                    ),
                                    Argument(
                                        "--endpoint", type=str, required=False,
                                        help="An Amazon S3 compatible endpoint."
                                    ),
                                ] + import_job_create_common_args,
                                "resolver": import_jobs_create_from_s3,
                            },
                            "from-azure-blob-storage": {
                                "help": "Create a data import job on the specified "
                                        "cluster from an Azure blob storage location.",
                                "extra_args": [
                                    # Type Azure Blob Storage params
                                    Argument(
                                        "--container-name", type=str,
                                        required=True,
                                        help="The name of the storage container "
                                             "where the file to be imported is located."
                                    ),
                                    Argument(
                                        "--blob-name", type=str, required=True,
                                        help="The absolute path in the storage "
                                             "container that points to the file to be "
                                             "imported. Globbing (use of *) is allowed."
                                    ),
                                    Argument(
                                        "--secret-id", type=str, required=True,
                                        help="The secret that contains the access key "
                                             "and secret key needed to access the file "
                                             "to be imported."
                                    ),
                                ] + import_job_create_common_args,
                                "resolver": import_jobs_create_from_azure_blob_storage,
                            },
                        },
                    },
                    "progress": {
                        "help": "Shows the progress of an import job.",
                        "extra_args": [
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The cluster the import jobs belong to."
                            ),
                            Argument(
                                "--import-job-id", type=str,
                                required=True,
                                help="The ID of the Import Job."
                            ),
                            Argument(
                                "--limit", type=str, required=False,
                                help="The number of files returned."
                                     "Use keywork 'ALL' to have no limit applied."
                            ),
                            Argument(
                                "--offset", type=int, required=False,
                                help="The offset to skip before beginning to "
                                     "return the files."
                            ),
                            Argument(
                                "--summary", type=lambda x: bool(strtobool(str(x))),
                                required=False,
                                help="Show only global progress."
                            ),
                        ],
                        "resolver": import_job_progress,
                    },
                },
            },
            "export-jobs": {
                "help": "Manage data export jobs.",
                "commands": {
                    "delete": {
                        "help": "Cancels an already running data export job that has "
                                "not finished yet. "
                                "If the job has already finished it deletes it from "
                                "the job history.",
                        "extra_args": [
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The cluster the job belongs to."
                            ),
                            Argument(
                                "--export-job-id", type=str,
                                required=True,
                                help="The ID of the export job."
                            ),
                        ],
                        "resolver": export_jobs_delete,
                    },
                    "list": {
                        "help": "Lists data export jobs that belong to a specific "
                                "cluster.",
                        "extra_args": [
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The cluster the export jobs belong to."
                            ),
                        ],
                        "resolver": export_jobs_list,
                    },
                    "create": {
                        "help": "Create a data export job for the specified cluster.",
                        "extra_args": [
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The cluster the data will be exported from."
                            ),
                            Argument(
                                "--table",
                                type=str,
                                required=True,
                                help="The table the data will be exported from.",
                            ),
                            Argument(
                                "--file-format",
                                type=str,
                                required=True,
                                choices=["csv", "json", "parquet"],
                                help="The format of the data in the file.",
                            ),
                            Argument(
                                "--compression",
                                type=str,
                                required=False,
                                choices=["gzip"],
                                help="The compression method of the exported file.",
                            ),
                            Argument(
                                "--save-as",
                                type=str,
                                required=False,
                                help="The file on your local filesystem the data will "
                                     "be exported to. If not specified, you will "
                                     "receive the URL to download the file.",
                            ),
                        ],
                        "resolver": export_jobs_create,
                    },
                },
            },
            "scheduled-jobs": {
                "help": "Manage your scheduled sql jobs.",
                "commands": {
                    "create": {
                        "help": "Create a scheduled sql job to run at "
                                "specific times.",
                        "extra_args": [
                            Argument(
                                "--name", type=str, required=True,
                                help="Name of the sql job."
                            ),
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="Cluster where the job should be run."
                            ),
                            Argument(
                                "--cron", type=str, required=True,
                                help="Cron schedule of the sql job."
                            ),
                            Argument(
                                "--sql", type=str, required=True,
                                help="The sql statement the job should run."
                            ),
                            Argument(
                                "--enabled", type=str, required=True,
                                help="Enable or disable the job."
                            )
                        ],
                        "resolver": create_scheduled_job,
                    },
                    "list": {
                        "help": "Get all scheduled sql jobs.",
                        "extra_args": [
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The cluster of which jobs should be listed."
                            )
                        ],
                        "resolver": get_scheduled_jobs,
                    },
                    "logs": {
                        "help": "Logs of a scheduled sql job.",
                        "extra_args": [
                            Argument(
                                "--job-id", type=str, required=True,
                                help="The job id of the job log to be listed."
                            ),
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The cluster of which the job log "
                                     "should be listed."
                            )
                        ],
                        "resolver": get_scheduled_job_log,
                    },
                    "delete": {
                        "help": "Delete specified scheduled sql job.",
                        "extra_args": [
                            Argument(
                                "--job-id", type=str, required=True,
                                help="The job id of the job to be deleted."
                            ),
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The cluster of which the job "
                                     "should be deleted."
                            ),
                        ],
                        "resolver": delete_scheduled_job,
                    },
                    "edit": {
                        "help": "Edit specified scheduled sql job.",
                        "extra_args": [
                            Argument(
                                "--job-id", type=str, required=True,
                                help="The id of the job to edit."
                            ),
                            Argument(
                                "--cluster-id", type=str, required=True,
                                help="The cluster id where the job was created."
                            ),
                            Argument(
                                "--name", type=str, required=True,
                                help="The name of the sql job."
                            ),
                            Argument(
                                "--sql", type=str, required=True,
                                help="The sql statement of the sql job."
                            ),
                            Argument(
                                "--cron", type=str, required=True,
                                help="Cron schedule of the sql job."
                            ),
                            Argument(
                                "--enabled", type=str, required=True,
                                help="Enable or disable the sql job."
                            ),
                        ],
                        "resolver": edit_scheduled_job,
                    }
                }
            },
        },
    },
    "products": {
        "help": "Manage Products.",
        "commands": {
            "list": {
                "help": "List all available products in the current region.",
                "extra_args": [
                    Argument(
                        "--kind", type=str, required=False, help="The product kind."
                    ),
                ],
                "resolver": products_list,
            },
        },
    },
    "organizations": {
        "help": "Manage organizations.",
        "commands": {
            "create": {
                "help": "Create a new organization.",
                "extra_args": [
                    Argument(
                        "--name", type=str, required=True,
                        help="The organization name to use.",
                    ),
                    Argument(
                        "--plan-type", type=int, required=False,
                        choices=[1, 2, 3, 4, 5, 6],
                        help="The support plan to use for the organization. Argument "
                             "is for superusers only.",
                    ),
                ],
                "resolver": organizations_create,
            },
            "get": {
                "help": (
                    "Get an organization by its ID."
                ),
                "extra_args": [
                    Argument(
                        "id", type=str,
                        help="The ID of the organization.",
                    ),
                ],
                "resolver": organizations_get,
            },
            "list": {
                "help": "List all organizations the current user has access to.",
                "resolver": organizations_list,
            },
            "edit": {
                "help": "Edit the specified organization.",
                "extra_args": [
                    Argument(
                        "--name", type=str, required=False,
                        help="The new organization name to use.",
                    ),
                    Argument(
                        "--plan-type", type=int, required=False,
                        choices=[1, 2, 3, 4, 5, 6],
                        help="The new support plan to use for the organization.",
                    ),
                    Argument(
                        "--org-id", type=str, required=False,
                        help="The organization ID to use.",
                    ),
                ],
                "resolver": organizations_edit,
            },
            "delete": {
                "help": "Delete the specified organization.",
                "extra_args": [
                    Argument("-y", "--yes", action="store_true", default=False),
                    Argument(
                        "--org-id", type=str, required=False,
                        help="The organization ID to use.",
                    ),
                ],
                "resolver": organizations_delete,
            },
            "auditlogs": {
                "help": "Show audit logs for an organization.",
                "commands": {
                    "list": {
                        "help": "List all audit events in the current organization.",
                        "extra_args": [
                            Argument(
                                "--action", type=str, required=False,
                                help="The audit event action.",
                            ),
                            Argument(
                                "--from", type=str, required=False, dest="from_",
                                help="Only show events from this point in time.",
                                metavar="FROM",
                            ),
                            Argument(
                                "--to", type=str, required=False,
                                help="Only show events older than this.",
                            ),
                            Argument(
                                "--org-id", type=str, required=False,
                                help="The organization ID to use.",
                            ),
                        ],
                        "resolver": auditlogs_list,
                    },
                },
            },
            "users": {
                "help": "Manage users in an organization.",
                "commands": {
                    "add": {
                        "help": "Add the selected user to an organization.",
                        "extra_args": [
                            Argument(
                                "--user", type=str, required=True,
                                help="The user email address or ID to use.",
                            ),
                            Argument(
                                "--role", type=str, required=True,
                                help=(
                                    "The role FQN to use. Run `croud users roles list` "
                                    "for a list of available roles."
                                ),
                            ),
                            Argument(
                                "--org-id", type=str, required=False,
                                help="The organization ID to use.",
                            ),
                        ],
                        "resolver": org_users_add,
                    },
                    "list": {
                        "help": "List all users within an organization.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=False,
                                help="The organization ID to use.",
                            ),
                        ],
                        "resolver": org_users_list,
                    },
                    "remove": {
                        "help": "Remove the selected user from an organization.",
                        "extra_args": [
                            Argument(
                                "--user", type=str, required=True,
                                help="The user email address or ID to use.",
                            ),
                            Argument(
                                "--org-id", type=str, required=False,
                                help="The organization ID to use.",
                            ),
                        ],
                        "resolver": org_users_remove,
                    },
                },
            },
            "secrets": {
                "help": "Manage organization's secrets.",
                "commands" : {
                    "list": {
                        "help": "Lists all the secrets that belong to an organization.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=True,
                                help="The organization ID to use.",
                            ),
                        ],
                        "resolver": org_secrets_list,
                    },
                    "create": {
                        "help": "Creates a new secret for the given organization.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=True,
                                help="The organization ID to use.",
                            ),
                            Argument(
                                "--name", type=str, required=True,
                                help="The name the secret will be known as.",
                            ),
                            Argument(
                                "--type", type=str, required=True,
                                choices=["AWS", "AZURE"],
                                help="The type of Secret. Either AWS or Azure.",
                            ),
                            # AWS arguments
                            Argument(
                                "--access-key", type=str, required=False,
                                help="For an AWS type secret, the access key ID.",
                            ),
                            Argument(
                                "--secret-key", type=str, required=False,
                                help="For an AWS type secret, the secret key.",
                            ),
                            # Azure arguments
                            Argument(
                                "--connection-string", type=str, required=False,
                                help="For an Azure type secret, the connection string "
                                     "or URL that grants access to a resource.",
                            ),
                        ],
                        "resolver": org_secrets_create,
                    },
                    "delete": {
                        "help": "Deletes the secret that matches the given ID for the "
                                "organization specified.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=True,
                                help="The organization ID to use.",
                            ),
                            Argument(
                                "--secret-id", type=str, required=True,
                                help="The secret ID to use.",
                            ),
                        ],
                        "resolver": org_secrets_delete,
                    },
                }
            },
            "files": {
                "help": "Manage organization's files.",
                "commands": {
                    "get": {
                        "help": (
                            "Get a file by its ID."
                        ),
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=True,
                                help="The organization ID to use.",
                            ),
                            Argument(
                                "--file-id", type=str, required=True,
                                help="The ID of the file.",
                            ),
                        ],
                        "resolver": org_files_get,
                    },
                    "list": {
                        "help": "List all files uploaded to this organization.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=True,
                                help="The organization ID to use.",
                            ),
                        ],
                        "resolver": org_files_list,
                    },
                    "create": {
                        "help": "Uploads a new file to the organization.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=True,
                                help="The organization ID to use.",
                            ),
                            Argument(
                                "--file-path", type=str, required=True,
                                help="The local file to be uploaded.",
                            ),
                            Argument(
                                "--name", type=str, required=False,
                                help="The name that will be displayed when listing the "
                                     "file. If not provided, the file name will be "
                                     "used.",
                            ),
                        ],
                        "resolver": org_files_create,
                    },
                    "delete": {
                        "help": "Deletes a previously uploaded file.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=True,
                                help="The organization ID to use.",
                            ),
                            Argument(
                                "--file-id", type=str, required=True,
                                help="The ID of the file.",
                            ),
                        ],
                        "resolver": org_files_delete,
                    },
                }
            },
            "credits": {
                "help": "Manage organization's credits.",
                "commands" : {
                    "list": {
                        "help": "Lists all credits of an organization.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=True,
                                help="The organization ID to use.",
                            ),
                            Argument(
                                "--status", type=str, required=False,
                                help="Filter credits by status, comma separated. Valid "
                                     "values are ``ACTIVE`` and ``EXPIRED``. By "
                                     "default only ``ACTIVE`` credits are listed.",
                            ),
                        ],
                        "resolver": org_credits_list,
                    },
                    "create": {
                        "help": "Creates a new credit for the specified organization.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=True,
                                help="The organization ID to use.",
                            ),
                            Argument(
                                "--amount", type=float, required=True,
                                help="The amount to be credited in USD.",
                            ),
                            Argument(
                                "--expiration-date", type=str, required=True,
                                help="The expiration date of the credit in ISO 8601 "
                                     "format (i.e. ``2024-01-01T10:00:00Z``).",
                            ),
                            Argument(
                                "--comment", type=str, required=True,
                                help="The reason for creating this credit.",
                            ),
                        ],
                        "resolver": org_credits_create,
                    },
                    "edit": {
                        "help": "Edits the specified credit.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=True,
                                help="The organization ID to use.",
                            ),
                            Argument(
                                "--credit-id", type=str, required=True,
                                help="The credit ID to use.",
                            ),
                            Argument(
                                "--amount", type=float, required=False,
                                help="The amount to be credited in USD. It can only be "
                                     "increased.",
                            ),
                            Argument(
                                "--expiration-date", type=str, required=False,
                                help="The expiration date of the credit in ISO 8601 "
                                     "format (i.e. ``2024-01-01T10:00:00Z``).",
                            ),
                            Argument(
                                "--comment", type=str, required=False,
                                help="The reason for creating this credit.",
                            ),
                        ],
                        "resolver": org_credits_edit,
                    },
                    "expire": {
                        "help": "Expires the specified credit.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=True,
                                help="The organization ID to use.",
                            ),
                            Argument(
                                "--credit-id", type=str, required=True,
                                help="The credit ID to use.",
                            ),
                        ],
                        "resolver": org_credits_expire,
                    },
                }
            }
        },
    },
    "users": {
        "help": "Manage users.",
        "commands": {
            "list": {
                "help": "List all users.",
                "extra_args": [
                    Argument(
                        "--no-roles", action="store_true",
                        help="List users without roles.",
                    ),
                    Argument(
                        "--no-org", action="store_true",
                        help="Only show users that are not part of any organization.",
                    )
                ],
                "resolver": users_list,
            },
            "roles": {
                "help": "Manage user roles.",
                "commands": {
                    "list": {
                        "help": "List all available roles.",
                        "resolver": roles_list,
                    },
                },
            },
            "delete": {
                "help": "Delete the specified user. The command is for "
                        "superusers only.",
                "extra_args": [
                    Argument(
                        "--user-id", type=str, required=True,
                        help="The user ID to use.",
                    ),
                    Argument("-y", "--yes", action="store_true", default=False)
                ],
                "resolver": users_delete,
            },
        },
    },
    "api-keys": {
        "help": "Manage your own API keys.",
        "commands": {
            "list": {
                "help": "List all the API keys that belong to you.",
                "resolver": api_keys_list,
            },
            "create": {
                "help": "Create a new API key for your user. It will have the same "
                        "permissions as your user.",
                "resolver": api_keys_create,
            },
            "delete": {
                "help": "Deletes the API key specified that belongs to your user.",
                "resolver": api_keys_delete,
                "extra_args": [
                    Argument(
                        "--api-key", type=str, required=True,
                        help="The key that identifies an API key that belongs to your "
                             "user.",
                    ),
                ]
            },
            "edit": {
                "help": "Edit the API key specified that belongs to your user, "
                        "allowing you to activate or deactivate it.",
                "resolver": api_keys_edit,
                "extra_args": [
                    Argument(
                        "--api-key", type=str, required=True,
                        help="The key that identifies an API key that belongs to your "
                             "user.",
                    ),
                    Argument(
                        "--active", type=str, required=True,
                        choices=["true", "false"],
                        help="Either true or false. Determines whether the API key can "
                             "be used or not.",
                    ),
                ]
            },
        },
    },
    "regions": {
        "help": "Print information about available regions.",
        "commands": {
            "list": {
                "help": "List all available regions.",
                "extra_args": [
                    Argument(
                        "--org-id", type=str, required=False,
                        help="The organization ID to use.",
                    ),
                ],
                "resolver": regions_list,
            },
            "create": {
                "help": "Create a new region.",
                "resolver": regions_create,
                "extra_args": [
                    Argument(
                        "--aws-bucket", type=str, required=False,
                        help="The AWS S3 bucket where cluster backups will be stored.",
                    ),
                    Argument(
                        "--aws-region", type=str, required=False,
                        help="The AWS region where the S3 bucket for cluster"
                             " backups is expected to be.",
                    ),
                    Argument(
                        "--description", type=str, required=True,
                        help="The description of the new region.",
                    ),
                    Argument(
                        "--org-id", type=str, required=True,
                        help="The organization ID to use.",
                    ),
                    Argument("-y", "--yes", action="store_true", default=False),
                ],
            },
            "delete": {
                "help": "Delete an existing edge region.",
                "resolver": regions_delete,
                "extra_args": [
                    Argument(
                        "--name", type=str, required=True,
                        help="The name of the region that will be deleted."
                    ),
                    Argument("-y", "--yes", action="store_true", default=False),
                ],
            },
        }
    },
    "subscriptions": {
        "help": "Manage subscriptions.",
        "commands": {
            "create": {
                "help": ("Create a subscription in the specified organization. Command "
                         "is for superusers only."),
                "extra_args": [
                    Argument(
                        "--type", type=str, required=True,
                        choices=['contract'],
                        help="The subscription type to use.",
                    ),
                    Argument(
                        "--org-id", type=str, required=True,
                        help="The organization ID to use.",
                    ),
                ],
                "resolver": subscriptions_create,
            },
            "get": {
                "help": (
                    "Get a subscription by its ID."
                ),
                "extra_args": [
                    Argument(
                        "id", type=str,
                        help="The ID of the subscription.",
                    ),
                ],
                "resolver": subscriptions_get,
            },
            "delete": {
                "help": "For Stripe or contract subscriptions only. Cancels the "
                        "specified subscription. "
                        "CAVEAT EMPTOR! "
                        "This will delete any clusters running in this subscription.",
                "extra_args": [
                    Argument(
                        "--subscription-id", type=str, required=True,
                        help="The ID of the subscription.",
                    ),
                    Argument("-y", "--yes", action="store_true", default=False),
                ],
                "resolver": subscription_delete,
            },
            "list": {
                "help": "List all subscriptions the current user has access to.",
                "extra_args": [
                    Argument(
                        "--org-id", type=str, required=False,
                        help="The organization ID to use.",
                    ),
                ],
                "resolver": subscriptions_list,
            },
        },
    },
    "cloud-configurations": {
        "help": "Manage configurations of CrateDB Cloud.",
        "commands": {
            "set": {
                "help": (
                    "Set a configuration key/value. This command "
                    "is for superusers only."
                ),
                "extra_args": [
                    Argument(
                        "--key", type=str, required=True,
                        help="The configuration key to set.",
                    ),
                    Argument(
                        "--value", type=str, required=True,
                        help="The configuration value to use.",
                    ),
                    Argument(
                        "--org-id", type=str, required=False,
                        help="Override the value for a single organization only.",
                    ),
                ],
                "resolver": cloud_configurations_set,
            },
            "get": {
                "help": (
                    "Get the configuration value by its key. This command "
                    "is for superusers only."
                ),
                "extra_args": [
                    Argument(
                        "--key", type=str, required=True,
                        help="The configuration key to get.",
                    ),
                    Argument(
                        "--org-id", type=str, required=False,
                        help="Optionally get the value for a certain organization. "
                             "Defaults to the global configuration value.",
                    ),
                ],
                "resolver": cloud_configurations_get,
            },
            "list": {
                "help": (
                    "List all cloud configurations. This command "
                    "is for superusers only."
                ),
                "extra_args": [
                    Argument(
                        "--org-id", type=str, required=False,
                        help="Optionally get the values for a certain organization. "
                             "Defaults to the global configuration values.",
                    ),
                ],
                "resolver": cloud_configurations_list,
            }
        },
    }
}
# fmt: on


def get_parser():
    tree = {
        "help": "A command line interface for CrateDB Cloud.",
        "commands": command_tree,
    }
    return create_parser(tree)


def main():
    if not CONFIG.is_valid():
        print_error(
            "Your configuration file is incompatible with the current version of croud."
        )
        print_info(
            f"Please delete the file '{CONFIG._file_path}' or update it manually."
        )
        sys.exit(1)

    colorama.init()

    parser = get_parser()
    shtab.add_argument_to(parser)  # Tab completion stuff
    params = parser.parse_args(sys.argv[1:])
    if "resolver" in params:
        fn = params.resolver
        del params.resolver
        with HALO:
            fn(params)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
import os
from argparse import Namespace
from typing import Any, Tuple

import bitmath
import requests
from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from croud.api import Client
from croud.config import CONFIG, get_output_format
from croud.printer import print_error, print_info, print_response
from croud.tools.spinner import HALO
from croud.util import org_id_config_fallback, require_confirmation


def organizations_create(args: Namespace) -> None:
    client = Client.from_args(args)
    if args.plan_type:
        body = {"name": args.name, "plan_type": args.plan_type}
    else:
        body = {"name": args.name}

    data, errors = client.post("/api/v2/organizations/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "plan_type"],
        success_message="Organization created.",
        output_fmt=get_output_format(args),
    )


@org_id_config_fallback
def organizations_edit(args: Namespace) -> None:
    client = Client.from_args(args)
    body = {}
    if args.plan_type:
        body["plan_type"] = args.plan_type
    if args.name:
        body["name"] = args.name
    if not body:
        print_error("No input arguments found.")
        exit(1)

    data, errors = client.put(f"/api/v2/organizations/{args.org_id}/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "plan_type"],
        success_message="Organization edited.",
        output_fmt=get_output_format(args),
    )


def organizations_get(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/organizations/{args.id}/")
    print_response(
        data=data,
        errors=errors,
        output_fmt=get_output_format(args),
    )


def organizations_list(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get("/api/v2/organizations/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "plan_type"],
        output_fmt=get_output_format(args),
    )


@org_id_config_fallback
@require_confirmation(
    "Are you sure you want to delete the organization?",
    cancel_msg="Organization deletion cancelled.",
)
def organizations_delete(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(f"/api/v2/organizations/{args.org_id}/")
    print_response(
        data=data,
        errors=errors,
        success_message="Organization deleted.",
        output_fmt=get_output_format(args),
    )

    if errors is None and args.org_id == CONFIG.organization:
        CONFIG.set_organization_id(CONFIG.name, None)


def org_files_list(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/organizations/{args.org_id}/files/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "status"],
        output_fmt=get_output_format(args),
    )


def op_upload_file_to_org(
    client, org_id: str, file_path: str, file_name: str = None
) -> Tuple[Any, Any]:
    if not os.path.isfile(file_path):
        return None, {"message": "The file path does not exist."}

    # Name the file in Cloud. If no name is provided the file path will be used.
    payload = {"name": file_name or file_path}
    print_info("Creating a new file upload...")
    data, errors = client.post(f"/api/v2/organizations/{org_id}/files/", body=payload)

    # HALO spinner needs to be stopped to display the progress bar
    HALO.stop()
    if not errors and data and data.get("upload_url"):
        # Progress bar works by wrapping the file and monitoring its read ops
        with tqdm(
            total=os.path.getsize(file_path),
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as t:
            with open(file_path, "rb") as file_upload:
                print_info("Uploading the file...")
                wrapped_file = CallbackIOWrapper(t.update, file_upload, "read")
                resp = requests.put(data["upload_url"], data=wrapped_file)
                if resp.status_code < 200 or resp.status_code >= 300:
                    errors = {
                        "code": resp.status_code,
                        "message": "There was an error while trying to upload the file",
                        "upload_url": data["upload_url"],
                    }

    return data, errors


def org_secrets_create(args: Namespace) -> None:
    client = Client.from_args(args)
    payload = {
        "name": args.name,
        "type": args.type,
    }

    if args.type == "AWS":
        if not args.access_key or not args.secret_key:
            print_error(
                "Both access_key and secret_key are required for secret type AWS."
            )
            return
        payload["data"] = {
            "access_key": args.access_key,
            "secret_key": args.secret_key,
        }
    elif args.type == "AZURE":
        if not args.connection_string:
            print_error("Argument connection-string is required for secret type Azure.")
            return
        payload["data"] = {
            "azure_secret": {
                "connection_string": args.connection_string,
            }
        }

    data, errors = client.post(
        f"/api/v2/organizations/{args.org_id}/secrets/", body=payload
    )
    print_response(
        data=data,
        errors=errors,
        keys=["id", "type", "name", "description"],
        success_message="Secret created.",
        output_fmt=get_output_format(args),
    )


def org_secrets_list(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/organizations/{args.org_id}/secrets/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "type", "name", "description"],
        output_fmt=get_output_format(args),
    )


def org_secrets_delete(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(
        f"/api/v2/organizations/{args.org_id}/secrets/{args.secret_id}/"
    )
    print_response(
        data=data,
        errors=errors,
        keys=["id", "type", "name", "description"],
        success_message="Secret deleted.",
        output_fmt=get_output_format(args),
    )


def org_files_create(args: Namespace) -> None:
    client = Client.from_args(args)

    data, errors = op_upload_file_to_org(client, args.org_id, args.file_path, args.name)
    if errors or not data:
        print_response(
            data=data,
            errors=errors,
            success_message="There was a problem uploading the file.",
            keys=["id", "name", "status"],
            output_fmt=get_output_format(args),
        )
    else:
        # Refresh the data from the file resource to get the latest status.
        data_refresh, errors = client.get(f"/api/v2/organizations/{args.org_id}/files/")
        if data_refresh:
            file = [x for x in data_refresh if x["id"] == data["id"]]
        print_response(
            data=file,
            errors=errors,
            success_message="File upload completed!",
            keys=["id", "name", "status"],
            output_fmt=get_output_format(args),
        )


def org_files_delete(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(
        f"/api/v2/organizations/{args.org_id}/files/{args.file_id}/"
    )
    print_response(
        data=data,
        errors=errors,
        success_message="File upload deleted.",
        output_fmt=get_output_format(args),
    )


def org_files_get(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(
        f"/api/v2/organizations/{args.org_id}/files/{args.file_id}/"
    )
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "status", "file_size", "download_url"],
        output_fmt=get_output_format(args),
        transforms={
            "file_size": _transform_file_size,
        },
    )


def _transform_file_size(size_bytes):
    if not size_bytes:
        return None
    return bitmath.Byte(size_bytes).best_prefix().format("{value:.2f} {unit}")


def org_credits_create(args: Namespace) -> None:
    client = Client.from_args(args)
    payload = {
        "amount": int(args.amount * 100),
        "expiration_date": args.expiration_date,
        "comment": args.comment,
    }
    data, errors = client.post(
        f"/api/v2/organizations/{args.org_id}/credits/", body=payload
    )
    print_response(
        data=data,
        errors=errors,
        keys=[
            "id",
            "original_amount",
            "expiration_date",
            "comment",
            "status",
        ],
        success_message="Credit created.",
        output_fmt=get_output_format(args),
        transforms={
            "original_amount": _transform_credits_amount_to_usd,
        },
    )


def org_credits_edit(args: Namespace) -> None:
    client = Client.from_args(args)
    payload = {}
    if getattr(args, "amount") is not None:
        payload["amount"] = int(args.amount * 100)
    if args.expiration_date:
        payload["expiration_date"] = args.expiration_date
    if args.comment:
        payload["comment"] = args.comment
    if not payload:
        print_error("No input arguments found.")
        exit(1)

    data, errors = client.patch(
        f"/api/v2/organizations/{args.org_id}/credits/{args.credit_id}/", body=payload
    )
    print_response(
        data=data,
        errors=errors,
        keys=[
            "id",
            "original_amount",
            "expiration_date",
            "comment",
            "status",
        ],
        output_fmt=get_output_format(args),
        success_message="Credit edited.",
        transforms={
            "original_amount": _transform_credits_amount_to_usd,
        },
    )


def org_credits_list(args: Namespace) -> None:
    client = Client.from_args(args)
    options = {}
    if args.status:
        options["status"] = args.status
    data, errors = client.get(
        f"/api/v2/organizations/{args.org_id}/credits/", params=options
    )
    print_response(
        data=data,
        errors=errors,
        keys=[
            "id",
            "original_amount",
            "remaining_amount",
            "expiration_date",
            "comment",
            "status",
        ],
        output_fmt=get_output_format(args),
        transforms={
            "original_amount": _transform_credits_amount_to_usd,
            "remaining_amount": _transform_credits_amount_to_usd,
        },
    )


def _transform_credits_amount_to_usd(amount_cents):
    if not amount_cents:
        return 0
    return f"${round(amount_cents / 100, 2)}"


def org_credits_expire(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(
        f"/api/v2/organizations/{args.org_id}/credits/{args.credit_id}/"
    )
    print_response(
        data=data,
        errors=errors,
        success_message="Credit expired.",
        output_fmt=get_output_format(args),
    )



from unittest import mock

from croud.api import Client, RequestMethod
from tests.util import assert_rest, call_command, gen_uuid


@mock.patch.object(Client, "request", return_value=({}, None))
def test_create_job(mock_request):
    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.POST:
            return {
                "name": "test-job",
                "id": gen_uuid(),
                "cron": "1 1 * * *",
                "sql": "CREATE TABLE test (id TEXT)",
                "enable": True,
            }, None
        if args[0] == RequestMethod.GET and "/jwt/" in args[1]:
            return {
                "token": "xyz",
                "expiry": "01.02.2024",
            }, None
        if args[0] == RequestMethod.GET:
            return {"fqdn": "my.cluster.cloud", "name": "mycluster"}, None
        return None, None

    mock_request.side_effect = mock_call

    cluster_id = gen_uuid()
    call_command(
        "croud",
        "clusters",
        "scheduled-jobs",
        "create",
        "--name",
        "test-job",
        "--cluster-id",
        cluster_id,
        "--cron",
        "1 1 * * *",
        "--sql",
        "CREATE TABLE test (id TEXT)",
        "--enabled",
        "True",
    )

    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/jwt/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/scheduled-jobs/",
        body={
            "name": "test-job",
            "cron": "1 1 * * *",
            "sql": "CREATE TABLE test (id TEXT)",
            "enabled": "True",
        },
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_get_scheduled_jobs(mock_request):
    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/jwt/" in args[1]:
            return {
                "token": "xyz",
                "expiry": "01.02.2024",
            }, None
        if args[0] == RequestMethod.GET:
            return {"fqdn": "my.cluster.cloud", "name": "mycluster"}, None
        if args[0] == RequestMethod.GET and "/scheduled-jobs/" in args[1]:
            return {
                "name": "test-job",
                "id": gen_uuid(),
                "cron": "1 1 * * *",
                "sql": "CREATE TABLE test (id TEXT)",
                "enabled": True,
                "next_run_time": "02.02.2024",
            }, None
        return None, None

    mock_request.side_effect = mock_call

    cluster_id = gen_uuid()
    call_command(
        "croud",
        "clusters",
        "scheduled-jobs",
        "list",
        "--cluster-id",
        cluster_id,
    )

    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/jwt/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        "/api/scheduled-jobs/",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_get_scheduled_job_log(mock_request):
    job_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/jwt/" in args[1]:
            return {
                "token": "xyz",
                "expiry": "01.02.2024",
            }, None
        if args[0] == RequestMethod.GET:
            return {"fqdn": "my.cluster.cloud", "name": "mycluster"}, None
        if args[0] == RequestMethod.GET and "/scheduled-jobs/" in args[1]:
            return {
                "job_id": job_id,
                "start": "02.02.2024",
                "end": "03.02.2024",
                "error": None,
                "statements": "CREATE TABLE test (id TEXT)",
            }, None
        return None, None

    mock_request.side_effect = mock_call

    cluster_id = gen_uuid()
    call_command(
        "croud",
        "clusters",
        "scheduled-jobs",
        "logs",
        "--job-id",
        job_id,
        "--cluster-id",
        cluster_id,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/jwt/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/scheduled-jobs/{job_id}/log",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_delete_scheduled_job(mock_request):
    job_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/jwt/" in args[1]:
            return {
                "token": "xyz",
                "expiry": "01.02.2024",
            }, None
        if args[0] == RequestMethod.GET:
            return {"fqdn": "my.cluster.cloud", "name": "mycluster"}, None
        return None, None

    mock_request.side_effect = mock_call

    cluster_id = gen_uuid()
    call_command(
        "croud",
        "clusters",
        "scheduled-jobs",
        "delete",
        "--job-id",
        job_id,
        "--cluster-id",
        cluster_id,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/jwt/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.DELETE,
        f"/api/scheduled-jobs/{job_id}",
        any_times=True,
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_edit_scheduled_job(mock_request):
    job_id = gen_uuid()

    def mock_call(*args, **kwargs):
        if args[0] == RequestMethod.GET and "/jwt/" in args[1]:
            return {
                "token": "xyz",
                "expiry": "01.02.2024",
            }, None
        if args[0] == RequestMethod.GET:
            return {"fqdn": "my.cluster.cloud", "name": "mycluster"}, None
        if args[0] == RequestMethod.PUT:
            return {
                "name": "test-job-edit",
                "id": gen_uuid(),
                "cron": "2 2 * * *",
                "sql": "CREATE TABLE test (id TEXT)",
                "enabled": False,
                "next_run_time": "02.02.2024",
            }, None
        return None, None

    mock_request.side_effect = mock_call

    cluster_id = gen_uuid()
    call_command(
        "croud",
        "clusters",
        "scheduled-jobs",
        "edit",
        "--job-id",
        job_id,
        "--cluster-id",
        cluster_id,
        "--name",
        "test-job-edit",
        "--cron",
        "2 2 * * *",
        "--sql",
        "CREATE TABLE test (id TEXT)",
        "--enabled",
        "False",
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/jwt/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/clusters/{cluster_id}/",
        any_times=True,
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/scheduled-jobs/{job_id}",
        body={
            "name": "test-job-edit",
            "cron": "2 2 * * *",
            "sql": "CREATE TABLE test (id TEXT)",
            "enabled": "False",
        },
        any_times=True,
    )



from unittest import mock

import pytest

from croud.api import Client, RequestMethod
from tests.util import assert_rest, call_command


@mock.patch.object(Client, "request", return_value=({}, None))
def test_api_keys_list(mock_request):
    call_command("croud", "api-keys", "list")
    assert_rest(
        mock_request,
        RequestMethod.GET,
        "/api/v2/users/me/api-keys/",
        params=None,
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_api_keys_create(mock_request):
    call_command("croud", "api-keys", "create")
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/users/me/api-keys/",
        params=None,
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_api_keys_delete(mock_request):
    call_command("croud", "api-keys", "delete", "--api-key", "key-to-be-deleted")
    assert_rest(
        mock_request,
        RequestMethod.DELETE,
        "/api/v2/users/me/api-keys/key-to-be-deleted/",
        params=None,
    )


@mock.patch.object(Client, "request", return_value=({}, None))
@pytest.mark.parametrize("target_status", [True, False])
def test_api_keys_edit(mock_request, target_status):
    call_command(
        "croud",
        "api-keys",
        "edit",
        "--api-key",
        "target-key",
        "--active",
        str(target_status).lower(),
    )
    assert_rest(
        mock_request,
        RequestMethod.PATCH,
        "/api/v2/users/me/api-keys/target-key/",
        body={"active": target_status},
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import uuid
from datetime import datetime, timedelta
from unittest import mock

import pytest

from croud.api import Client, RequestMethod
from croud.organizations.users.commands import (
    role_fqn_transform as organization_role_fqn_transform,
)
from tests.util import assert_rest, call_command, gen_uuid

pytestmark = pytest.mark.usefixtures("config")

config_org_id = gen_uuid()
FALLBACK_ORG_ID_CONFIG: dict = {
    "auth": {
        "current_context": "prod",
        "contexts": {
            "prod": {"token": "", "organization_id": config_org_id},
            "dev": {"token": "", "organization_id": config_org_id},
            "local": {"token": "", "organization_id": config_org_id},
        },
    },
    "region": "bregenz.a1",
    "output_fmt": "table",
}


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_create(mock_request):
    call_command("croud", "organizations", "create", "--name", "test-org")
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/organizations/",
        body={"name": "test-org"},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_create_plan_type(mock_request):
    call_command(
        "croud", "organizations", "create", "--name", "test-org", "--plan-type", "3"
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/organizations/",
        body={"name": "test-org", "plan_type": 3},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_edit(mock_request):
    org_id = gen_uuid()
    call_command(
        "croud",
        "organizations",
        "edit",
        "--name",
        "new-org-name",
        "--plan-type",
        "3",
        "--org-id",
        org_id,
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/organizations/{org_id}/",
        body={"name": "new-org-name", "plan_type": 3},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_edit_name(mock_request):
    org_id = gen_uuid()
    call_command(
        "croud", "organizations", "edit", "--name", "new-org-name", "--org-id", org_id
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/organizations/{org_id}/",
        body={"name": "new-org-name"},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_edit_plan_type(mock_request):
    org_id = gen_uuid()
    call_command(
        "croud", "organizations", "edit", "--plan-type", "3", "--org-id", org_id
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/organizations/{org_id}/",
        body={"plan_type": 3},
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_organizations_edit_no_arguments(mock_request, capsys):
    org_id = gen_uuid()
    with pytest.raises(SystemExit):
        call_command("croud", "organizations", "edit", "--org-id", org_id)

    _, err = capsys.readouterr()
    assert "No input arguments found." in err


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_list(mock_request):
    call_command("croud", "organizations", "list")
    assert_rest(mock_request, RequestMethod.GET, "/api/v2/organizations/")


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_get(mock_request):
    id = str(uuid.uuid4())
    call_command("croud", "organizations", "get", id)
    assert_rest(mock_request, RequestMethod.GET, f"/api/v2/organizations/{id}/")


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_organizations_delete(mock_request, capsys):
    org_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["Y"]) as mock_input:
        call_command("croud", "organizations", "delete", "--org-id", org_id)
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/organizations/{org_id}/")
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the organization? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Organization deleted." in err_output


@mock.patch.object(Client, "request", return_value=(None, {"errors": "Message"}))
def test_organizations_delete_failure_org_id_not_deleted_from_config(
    mock_request, capsys, config
):
    org_id = gen_uuid()
    config.set_organization_id(config.name, org_id)

    with mock.patch("builtins.input", side_effect=["Y"]):
        call_command("croud", "organizations", "delete")

    assert_rest(
        mock_request,
        RequestMethod.DELETE,
        f"/api/v2/organizations/{org_id}/",
    )

    _, err = capsys.readouterr()
    assert err is not None
    assert config.organization == org_id


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_delete_org_id_from_local_config(
    mock_request,
    capsys,
    config,
):
    org_id = gen_uuid()
    config.set_organization_id(config.name, org_id)

    with mock.patch("builtins.input", side_effect=["Y"]):
        call_command("croud", "organizations", "delete")

    assert_rest(
        mock_request,
        RequestMethod.DELETE,
        f"/api/v2/organizations/{org_id}/",
    )

    assert config.organization is None


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_organizations_delete_flag(mock_request, capsys):
    org_id = gen_uuid()
    with mock.patch("builtins.input") as mock_input:
        call_command("croud", "organizations", "delete", "--org-id", org_id, "-y")
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/organizations/{org_id}/")
    mock_input.assert_not_called()

    _, err_output = capsys.readouterr()
    assert "Organization deleted." in err_output


@pytest.mark.parametrize("input", ["", "N", "No", "cancel"])
@mock.patch.object(Client, "request", return_value=(None, {}))
def test_organizations_delete_aborted(mock_request, capsys, input):
    org_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=[input]) as mock_input:
        call_command("croud", "organizations", "delete", "--org-id", org_id)
    mock_request.assert_not_called()
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the organization? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Organization deletion cancelled." in err_output


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_organizations_delete_aborted_with_input(mock_request, capsys):
    org_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["N"]) as mock_input:
        call_command("croud", "organizations", "delete", "--org-id", org_id)
    mock_request.assert_not_called()
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the organization? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Organization deletion cancelled." in err_output


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_auditlogs_list(mock_request):
    org_id = gen_uuid()

    call_command("croud", "organizations", "auditlogs", "list", "--org-id", org_id)
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/organizations/{org_id}/auditlogs/",
        params={},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_auditlogs_list_filtered(mock_request):
    org_id = gen_uuid()

    call_command(
        "croud",
        "organizations",
        "auditlogs",
        "list",
        "--org-id",
        org_id,
        "--action",
        "organization.create",
        "--from",
        "2019-10-11T12:13:14",
        "--to",
        "2019-11-12T12:34:56",
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/organizations/{org_id}/auditlogs/",
        params={
            "action": "organization.create",
            "from": "2019-10-11T12:13:14",
            "to": "2019-11-12T12:34:56",
        },
    )


@mock.patch.object(Client, "request", side_effect=[([{"id": 123}], None), ([], None)])
def test_organizations_auditlogs_list_pagination(mock_request: mock.Mock):
    org_id = gen_uuid()

    call_command("croud", "organizations", "auditlogs", "list", "--org-id", org_id)

    mock_request.call_args_list[0].assert_called_with(
        RequestMethod.GET,
        f"/api/v2/organizations/{org_id}/auditlogs/",
        params={},
    )
    mock_request.call_args_list[1].assert_called_with(
        RequestMethod.GET,
        f"/api/v2/organizations/{org_id}/auditlogs/",
        params={"last": 123},
    )


@pytest.mark.parametrize(
    "added,message",
    [(True, "User added to organization."), (False, "Role altered for user.")],
)
@mock.patch.object(Client, "request")
def test_organizations_users_add(mock_request, added, message, capsys):
    mock_request.return_value = ({"added": added}, None)

    org_id = gen_uuid()
    user = "test@crate.io"
    role_fqn = "org_admin"

    call_command(
        "croud",
        "organizations",
        "users",
        "add",
        "--user",
        user,
        "--org-id",
        org_id,
        "--role",
        role_fqn,
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        f"/api/v2/organizations/{org_id}/users/",
        body={"user": user, "role_fqn": role_fqn},
    )

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert message in err_output


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_users_list(mock_request):
    org_id = gen_uuid()

    call_command("croud", "organizations", "users", "list", "--org-id", org_id)
    assert_rest(
        mock_request, RequestMethod.GET, f"/api/v2/organizations/{org_id}/users/"
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_users_remove(mock_request):
    org_id = gen_uuid()
    user = "test@crate.io"

    call_command(
        "croud", "organizations", "users", "remove", "--user", user, "--org-id", org_id
    )
    assert_rest(
        mock_request,
        RequestMethod.DELETE,
        f"/api/v2/organizations/{org_id}/users/{user}/",
    )


def test_role_fqn_transform():
    user = {
        "organization_roles": [
            {"organization_id": "org-1", "role_fqn": "organization_admin"},
            {"organization_id": "org-2", "role_fqn": "organization_member"},
            {"organization_id": "org-3", "role_fqn": "organization_member"},
        ]
    }
    response = organization_role_fqn_transform(user["organization_roles"])
    assert response == "organization_admin"


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_files_get(mock_request):
    org_id = gen_uuid()
    file_id = gen_uuid()
    call_command(
        "croud",
        "organizations",
        "files",
        "get",
        "--org-id",
        org_id,
        "--file-id",
        file_id,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/organizations/{org_id}/files/{file_id}/",
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_files_list(mock_request):
    org_id = gen_uuid()

    call_command("croud", "organizations", "files", "list", "--org-id", org_id)
    assert_rest(
        mock_request, RequestMethod.GET, f"/api/v2/organizations/{org_id}/files/"
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_files_delete(mock_request):
    org_id = gen_uuid()
    file_id = gen_uuid()

    call_command(
        "croud",
        "organizations",
        "files",
        "delete",
        "--org-id",
        org_id,
        "--file-id",
        file_id,
    )
    assert_rest(
        mock_request,
        RequestMethod.DELETE,
        f"/api/v2/organizations/{org_id}/files/{file_id}/",
    )


@mock.patch("croud.organizations.commands.os.path.isfile", return_value=True)
@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_files_create(mock_request, mock_isfile):
    org_id = gen_uuid()
    file_path = "/path/to/file.json.gz"
    file_name = "my-file"
    call_command(
        "croud",
        "organizations",
        "files",
        "create",
        "--org-id",
        org_id,
        "--file-path",
        file_path,
        "--name",
        file_name,
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        f"/api/v2/organizations/{org_id}/files/",
        body={"name": file_name},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_secrets_list(mock_request):
    org_id = gen_uuid()

    call_command("croud", "organizations", "secrets", "list", "--org-id", org_id)
    assert_rest(
        mock_request, RequestMethod.GET, f"/api/v2/organizations/{org_id}/secrets/"
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_secrets_aws_create(mock_request):
    org_id = gen_uuid()
    name = "my_secret"
    secret_type = "AWS"
    access_key = "my_access_key"
    secret_key = "my_secret_key"

    call_command(
        "croud",
        "organizations",
        "secrets",
        "create",
        "--org-id",
        org_id,
        "--name",
        name,
        "--type",
        secret_type,
        "--access-key",
        access_key,
        "--secret-key",
        secret_key,
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        f"/api/v2/organizations/{org_id}/secrets/",
        body={
            "name": name,
            "type": secret_type,
            "data": {"access_key": access_key, "secret_key": secret_key},
        },
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_secrets_azure_create(mock_request):
    org_id = gen_uuid()
    name = "my_secret"
    secret_type = "AZURE"
    connection_string = "https://my-storage-account.blob.core.windows.net/my-container?my-auth-params"  # noqa

    call_command(
        "croud",
        "organizations",
        "secrets",
        "create",
        "--org-id",
        org_id,
        "--name",
        name,
        "--type",
        secret_type,
        "--connection-string",
        connection_string,
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        f"/api/v2/organizations/{org_id}/secrets/",
        body={
            "name": name,
            "type": secret_type,
            "data": {
                "azure_secret": {"connection_string": connection_string},
            },
        },
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_secrets_delete(mock_request):
    org_id = gen_uuid()

    call_command(
        "croud",
        "organizations",
        "secrets",
        "delete",
        "--org-id",
        org_id,
        "--secret-id",
        "my_secret_id",
    )
    assert_rest(
        mock_request,
        RequestMethod.DELETE,
        f"/api/v2/organizations/{org_id}/secrets/my_secret_id/",
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_credits_list(mock_request):
    org_id = gen_uuid()

    call_command("croud", "organizations", "credits", "list", "--org-id", org_id)
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/organizations/{org_id}/credits/",
        params={},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_credits_list_filter_status(mock_request):
    org_id = gen_uuid()

    call_command(
        "croud",
        "organizations",
        "credits",
        "list",
        "--org-id",
        org_id,
        "--status",
        "ACTIVE,EXPIRED",
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/organizations/{org_id}/credits/",
        params={"status": "ACTIVE,EXPIRED"},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_credits_create(mock_request):
    org_id = gen_uuid()
    amount = 123.45
    expiration_date = datetime.utcnow() + timedelta(days=30)
    comment = "Free Trial"

    call_command(
        "croud",
        "organizations",
        "credits",
        "create",
        "--org-id",
        org_id,
        "--amount",
        str(amount),
        "--expiration-date",
        expiration_date.isoformat(),
        "--comment",
        comment,
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        f"/api/v2/organizations/{org_id}/credits/",
        body={
            "amount": amount * 100,
            "expiration_date": expiration_date.isoformat(),
            "comment": comment,
        },
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_credits_expire(mock_request):
    org_id = gen_uuid()

    call_command(
        "croud",
        "organizations",
        "credits",
        "expire",
        "--org-id",
        org_id,
        "--credit-id",
        "my_credit_id",
    )
    assert_rest(
        mock_request,
        RequestMethod.DELETE,
        f"/api/v2/organizations/{org_id}/credits/my_credit_id/",
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_credits_edit(mock_request):
    org_id = gen_uuid()
    expiration_date = datetime.utcnow() + timedelta(days=60)
    call_command(
        "croud",
        "organizations",
        "credits",
        "edit",
        "--org-id",
        org_id,
        "--credit-id",
        "my_credit_id",
        "--amount",
        str(234.56),
        "--expiration-date",
        expiration_date.isoformat(),
        "--comment",
        "Starter Package",
    )
    assert_rest(
        mock_request,
        RequestMethod.PATCH,
        f"/api/v2/organizations/{org_id}/credits/my_credit_id/",
        body={
            "amount": 23456,
            "expiration_date": expiration_date.isoformat(),
            "comment": "Starter Package",
        },
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_credits_edit_amount(mock_request):
    org_id = gen_uuid()
    call_command(
        "croud",
        "organizations",
        "credits",
        "edit",
        "--org-id",
        org_id,
        "--credit-id",
        "my_credit_id",
        "--amount",
        str(234.56),
    )
    assert_rest(
        mock_request,
        RequestMethod.PATCH,
        f"/api/v2/organizations/{org_id}/credits/my_credit_id/",
        body={"amount": 23456},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_credits_edit_expiration_date(mock_request):
    org_id = gen_uuid()
    expiration_date = datetime.utcnow() + timedelta(days=60)
    call_command(
        "croud",
        "organizations",
        "credits",
        "edit",
        "--org-id",
        org_id,
        "--credit-id",
        "my_credit_id",
        "--expiration-date",
        expiration_date.isoformat(),
    )
    assert_rest(
        mock_request,
        RequestMethod.PATCH,
        f"/api/v2/organizations/{org_id}/credits/my_credit_id/",
        body={"expiration_date": expiration_date.isoformat()},
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_organizations_credits_edit_no_arguments(mock_request, capsys):
    org_id = gen_uuid()
    with pytest.raises(SystemExit):
        call_command(
            "croud",
            "organizations",
            "credits",
            "edit",
            "--org-id",
            org_id,
            "--credit-id",
            "my_credit_id",
        )

    _, err = capsys.readouterr()
    assert "No input arguments found." in err



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.api import Client
from croud.config import get_output_format
from croud.printer import print_response
from croud.util import require_confirmation


def subscriptions_create(args: Namespace) -> None:
    body = {"type": args.type, "organization_id": args.org_id}
    client = Client.from_args(args)
    data, errors = client.post("/api/v2/subscriptions/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "state", "provider"],
        success_message="Subscription created.",
        output_fmt=get_output_format(args),
    )


def subscriptions_get(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/subscriptions/{args.id}/")
    print_response(
        data=data,
        errors=errors,
        output_fmt=get_output_format(args),
    )


def subscriptions_list(args: Namespace) -> None:
    client = Client.from_args(args)
    if args.org_id:
        url = f"/api/v2/organizations/{args.org_id}/subscriptions/"
    else:
        url = "/api/v2/subscriptions/"
    data, errors = client.get(url)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "organization_id", "state", "provider"],
        output_fmt=get_output_format(args),
    )


@require_confirmation(
    "Are you sure you want to cancel this subscription? "
    "This will delete any clusters running in this subscription.",
    cancel_msg="Deletion cancelled.",
)
def subscription_delete(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(f"/api/v2/subscriptions/{args.subscription_id}/")
    print_response(
        data=data,
        errors=errors,
        success_message="Subscription cancelled.",
        keys=["id", "name", "organization_id", "state", "provider"],
        output_fmt=get_output_format(args),
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
import functools
import pathlib
import time
from argparse import Namespace
from datetime import datetime, timedelta, timezone
from shutil import copyfileobj
from typing import Any, Dict, Optional, cast

import bitmath
import requests
from tqdm.auto import tqdm
from yarl import URL

from croud.api import Client
from croud.clusters.exceptions import AsyncOperationNotFound
from croud.config import CONFIG, get_output_format
from croud.organizations.commands import op_upload_file_to_org
from croud.printer import print_error, print_info, print_response, print_success
from croud.tools.spinner import HALO
from croud.util import grand_central_jwt_token, require_confirmation


def clusters_get(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/clusters/{args.id}/")
    print_response(
        data=data,
        errors=errors,
        output_fmt=get_output_format(args),
    )


def clusters_list(args: Namespace) -> None:
    if args.org_id:
        url = f"/api/v2/organizations/{args.org_id}/clusters/"
    else:
        url = "/api/v2/clusters/"

    params = {}
    if args.project_id:
        params["project_id"] = args.project_id

    client = Client.from_args(args)
    data, errors = client.get(url, params=params)
    print_response(
        data=data,
        errors=errors,
        keys=[
            "id",
            "name",
            "num_nodes",
            "crate_version",
            "project_id",
            "username",
            "suspended",
            "fqdn",
            "channel",
        ],
        output_fmt=get_output_format(args),
    )


def clusters_deploy(args: Namespace) -> None:
    if not args.project_id and not (args.org_id and args.region):
        # We cannot have dynamically required params in croud, so have to verify here.
        print_error("Either a project id or organization id and region are required.")
        return

    body = {
        "cluster": {
            "crate_version": args.version,
            "name": args.cluster_name,
            "password": args.password,
            "product_name": args.product_name,
            "product_tier": args.tier,
            "username": args.username,
            "channel": args.channel,
        },
        "subscription_id": args.subscription_id,
    }
    if args.project_id:
        body["project_id"] = args.project_id
    else:
        body["project"] = {"name": args.cluster_name, "region": args.region}

    if args.unit:
        body["cluster"]["product_unit"] = args.unit
    _handle_edge_params(body["cluster"], args)

    client = Client.from_args(args)

    org_id = args.org_id or _lookup_organization_id_for_project(
        client, args, args.project_id
    )
    if not org_id:
        return

    data, errors = client.post(f"/api/v2/organizations/{org_id}/clusters/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "fqdn", "url"],
        output_fmt=get_output_format(args),
    )

    if errors or not data:
        return

    print_info("Cluster creation initiated. It may take a few minutes to complete.")

    _wait_for_completed_operation(
        client=client,
        cluster_id=data["id"],
        request_params={"type": "CREATE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{data['id']}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "fqdn", "url"],
        output_fmt=get_output_format(args),
    )


def clusters_scale(args: Namespace) -> None:
    body = {"product_unit": args.unit}
    client = Client.from_args(args)
    data, errors = client.put(f"/api/v2/clusters/{args.cluster_id}/scale/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "num_nodes"],
        output_fmt=get_output_format(args),
    )

    if errors:
        return

    print_info(
        "Cluster scaling initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "SCALE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "num_nodes"],
        output_fmt=get_output_format(args),
    )


def import_jobs_create_from_url(args: Namespace) -> None:
    extra_body = {
        "url": {
            "url": args.url,
        }
    }
    args.type = "url"
    import_jobs_create(args, extra_payload=extra_body)


def import_jobs_create_from_s3(args: Namespace) -> None:
    extra_body = {
        "s3": {
            "bucket": args.bucket,
            "file_path": args.file_path,
            "secret_id": args.secret_id,
        }
    }
    if args.endpoint:
        extra_body["s3"]["endpoint"] = args.endpoint
    args.type = "s3"
    import_jobs_create(args, extra_payload=extra_body)


def import_jobs_create_from_azure_blob_storage(args: Namespace) -> None:
    extra_body = {
        "azureblob": {
            "container_name": args.container_name,
            "blob_name": args.blob_name,
            "secret_id": args.secret_id,
        }
    }
    args.type = "azureblob"
    import_jobs_create(args, extra_payload=extra_body)


def _get_org_id_from_cluster_id(client, cluster_id: str) -> Optional[str]:
    data, errors = client.get(f"/api/v2/clusters/{cluster_id}/")
    if errors or not data:
        return None

    project_id = data["project_id"]

    data, errors = client.get(f"/api/v2/projects/{project_id}/")
    if errors or not data:
        return None

    return data["organization_id"]


def import_jobs_create_from_file(args: Namespace) -> None:
    file_id = args.file_id

    if not args.file_path and not args.file_id:
        print_error("Please specify either --file-id or --file-path")
        return

    if args.file_path:
        client = Client.from_args(args)
        org_id = _get_org_id_from_cluster_id(client, args.cluster_id)
        if not org_id:
            print_error("Could not find the organization related to the cluster.")
            return

        data, errors = op_upload_file_to_org(client, org_id, args.file_path)
        if errors or not data:
            print_error("Aborted import job creation.")
            return
        file_id = data["id"]

    extra_body = {
        "file": {
            "id": file_id,
        }
    }
    args.type = "file"
    import_jobs_create(args, extra_payload=extra_body)


def import_jobs_create(args: Namespace, extra_payload: Dict[str, Any]) -> None:
    body = {
        "type": args.type,
        "format": args.file_format,
        "destination": {
            "table": args.table,
        },
    }

    if args.compression:
        body["compression"] = args.compression

    if args.create_table is not None:
        body["destination"]["create_table"] = args.create_table

    if args.transformations:
        body["schema"] = {"select": args.transformations}

    if extra_payload:
        body.update(extra_payload)

    client = Client.from_args(args)
    data, errors = client.post(
        f"/api/v2/clusters/{args.cluster_id}/import-jobs/", body=body
    )
    print_response(
        data=data,
        errors=errors,
        keys=["id", "cluster_id", "status"],
        output_fmt=get_output_format(args),
    )

    if data:
        import_job_id = data["id"]

        _wait_for_completed_operation(
            client=client,
            cluster_id=args.cluster_id,
            request_params={"import_job_id": import_job_id},
            operation_status_func=_get_import_job_operation_status,
            feedback_func=(
                _data_job_feedback_func,
                ("import",),
            ),
        )


def import_jobs_delete(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(
        f"/api/v2/clusters/{args.cluster_id}/import-jobs/{args.import_job_id}/"
    )
    print_response(
        data=data,
        errors=errors,
        keys=["id", "cluster_id", "status"],
        output_fmt=get_output_format(args),
    )


def import_jobs_list(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/import-jobs/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "cluster_id", "status", "type", "destination"],
        output_fmt=get_output_format(args),
        transforms={
            "destination": lambda field: field.get("table"),
        },
    )


def import_job_progress(args: Namespace) -> None:
    client = Client.from_args(args)

    params = {}
    if args.limit:
        params["limit"] = args.limit
    if args.offset:
        params["offset"] = args.offset

    url = (
        f"/api/v2/clusters/{args.cluster_id}/import-jobs/{args.import_job_id}/progress/"
    )
    data, errors = client.get(url, params=params)

    if args.summary:
        print_response(
            data=data.get("progress", {}) if data else {},
            errors=errors,
            keys=[
                "percent",
                "records",
                "failed_records",
                "total_records",
                "total_files",
            ],
            output_fmt=get_output_format(args),
        )
    else:
        print_response(
            data=data.get("progress", {}).get("files", []) if data else {},
            errors=errors,
            keys=[
                "name",
                "percent",
                "records",
                "failed_records",
                "total_records",
            ],
            output_fmt=get_output_format(args),
        )


def clusters_upgrade(args: Namespace) -> None:
    body = {"crate_version": args.version}
    client = Client.from_args(args)
    data, errors = client.put(f"/api/v2/clusters/{args.cluster_id}/upgrade/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "crate_version"],
        output_fmt=get_output_format(args),
    )

    if errors:
        return

    print_info(
        "Cluster upgrade initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "UPGRADE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "crate_version"],
        output_fmt=get_output_format(args),
    )


@require_confirmation(
    "Are you sure you want to delete the cluster?",
    cancel_msg="Cluster deletion cancelled.",
)
def clusters_delete(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        success_message="Cluster deleted.",
        output_fmt=get_output_format(args),
    )


def clusters_restart_node(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(
        f"/api/v2/clusters/{args.cluster_id}/nodes/{args.ordinal}"
    )
    print_response(
        data=data,
        errors=errors,
        keys=["code", "status"],
        success_message=(
            "Node restarted. It may take a few minutes to complete the changes."
        ),
        output_fmt=get_output_format(args),
    )


def clusters_set_deletion_protection(args: Namespace) -> None:
    body = {"deletion_protected": args.value}
    client = Client.from_args(args)
    data, errors = client.put(
        f"/api/v2/clusters/{args.cluster_id}/deletion-protection/", body=body
    )
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "deletion_protected"],
        success_message=("Cluster deletion protection status successfully updated"),
        output_fmt=get_output_format(args),
    )


def clusters_set_suspended(args: Namespace) -> None:
    body = {"suspended": args.value}
    client = Client.from_args(args)
    data, errors = client.put(f"/api/v2/clusters/{args.cluster_id}/suspend/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "suspended"],
        output_fmt=get_output_format(args),
    )

    if errors:
        return
    print_info(
        "Updating the cluster status initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "SUSPEND", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "suspended"],
        output_fmt=get_output_format(args),
    )


@require_confirmation("This will overwrite all existing CIDR restrictions. Continue?")
def clusters_set_ip_whitelist(args: Namespace) -> None:
    networks = args.net
    networks = networks.split(",")
    cidr = []

    for net in networks:
        if len(net) > 0:
            cidr.append({"cidr": net})

    body = {"ip_whitelist": cidr}

    client = Client.from_args(args)
    data, errors = client.put(
        f"/api/v2/clusters/{args.cluster_id}/ip-restrictions/", body=body
    )  # type: ignore
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "ip_whitelist"],
        output_fmt=get_output_format(args),
    )

    if errors:
        return
    print_info(
        "Updating the IP Network whitelist initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "ALLOWED_CIDR_UPDATE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "ip_whitelist"],
        output_fmt=get_output_format(args),
    )


def _disk_size_transform(field):
    disk_size = str(bitmath.Byte(field["disk_size_per_node_bytes"]).best_prefix())
    return f"Disk size: {disk_size}"


def clusters_expand_storage(args: Namespace) -> None:
    body = {
        "disk_size_per_node_bytes": args.disk_size_gb * 1024 * 1024 * 1024,
    }

    client = Client.from_args(args)
    data, errors = client.put(
        f"/api/v2/clusters/{args.cluster_id}/storage/", body=body
    )  # type: ignore
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "hardware_specs"],
        output_fmt=get_output_format(args),
        transforms={"hardware_specs": _disk_size_transform},
    )

    if errors:
        return
    print_info(
        "Cluster storage expansion initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "EXPAND_STORAGE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "hardware_specs"],
        output_fmt=get_output_format(args),
        transforms={"hardware_specs": _disk_size_transform},
    )


def clusters_set_product(args: Namespace) -> None:
    body = {
        "product_name": args.product_name,
    }

    client = Client.from_args(args)
    data, errors = client.put(
        f"/api/v2/clusters/{args.cluster_id}/product/", body=body
    )  # type: ignore
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "product_name"],
        output_fmt=get_output_format(args),
    )

    if errors:
        return
    print_info(
        "Changing the cluster product initiated. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "CHANGE_COMPUTE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "product_name"],
        output_fmt=get_output_format(args),
    )


def clusters_set_backup_schedule(args: Namespace) -> None:
    body = {
        "backup_hours": args.backup_hours,
    }

    client = Client.from_args(args)
    data, errors = client.put(
        f"/api/v2/clusters/{args.cluster_id}/backup-schedule/", body=body
    )  # type: ignore
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "backup_schedule"],
        output_fmt=get_output_format(args),
    )

    if errors:
        return
    print_info(
        "Changing the cluster backup schedule. "
        "It may take a few minutes to complete the changes."
    )

    _wait_for_completed_operation(
        client=client,
        cluster_id=args.cluster_id,
        request_params={"type": "BACKUP_SCHEDULE_UPDATE", "limit": 1},
    )

    # Re-fetch the cluster's info
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "backup_schedule"],
        output_fmt=get_output_format(args),
    )


def clusters_snapshots_list(args: Namespace) -> None:
    days = args.days_ago
    start = datetime.now(tz=timezone.utc) - timedelta(days=days)
    url = f"/api/v2/clusters/{args.cluster_id}/snapshots/"

    client = Client.from_args(args)
    data, errors = client.get(url, params={"start": start.isoformat()})
    if data is not None and len(data) == 0:
        print_info(
            f"Looks like there are no snapshots for the last {days} days. "
            "Use --days-ago to broaden your search."
        )
        return

    print_response(
        data=data,
        errors=errors,
        keys=[
            "created",
            "repository",
            "snapshot",
        ],
        output_fmt=get_output_format(args),
    )


def clusters_snapshots_restore(args: Namespace) -> None:
    url = f"/api/v2/clusters/{args.cluster_id}/snapshots/restore/"

    body = {
        "repository": args.repository,
        "snapshot": args.snapshot,
    }
    if args.type:
        body["type"] = args.type
    else:
        # fallback for backwards compatibility
        body["type"] = "tables"
    if args.source_cluster_id:
        body["source_cluster_id"] = args.source_cluster_id
    if args.tables:
        tables = args.tables.strip().split(",")
        body["tables"] = [x.strip() for x in tables]
    if args.sections:
        sections = args.sections.strip().split(",")
        body["sections"] = [s.strip() for s in sections]

    client = Client.from_args(args)
    data, errors = client.post(url, body=body)

    if not errors:
        print_info(
            "Restoring the snapshot. Depending on the amount of data you have, this"
            " might take a very long time."
        )

        _wait_for_completed_operation(
            client=client,
            cluster_id=args.cluster_id,
            request_params={"type": "RESTORE_SNAPSHOT", "limit": 1},
        )

    print_response(
        data=data,
        errors=errors,
        keys=[
            "created",
            "repository",
            "snapshot",
        ],
        output_fmt=get_output_format(args),
    )


def export_jobs_create(args: Namespace) -> None:
    body = {
        "source": {
            "table": args.table,
        },
        "destination": {"format": args.file_format},
    }

    if args.compression:
        body["compression"] = args.compression

    client = Client.from_args(args)
    data, errors = client.post(
        f"/api/v2/clusters/{args.cluster_id}/export-jobs/", body=body
    )
    print_response(
        data=data,
        errors=errors,
        keys=["id", "cluster_id", "status"],
        output_fmt=get_output_format(args),
    )

    if data:
        export_job_id = data["id"]

        _wait_for_completed_operation(
            client=client,
            cluster_id=args.cluster_id,
            request_params={"export_job_id": export_job_id},
            operation_status_func=_get_export_job_operation_status,
            feedback_func=(
                _data_job_feedback_func,
                ("export",),
            ),
            post_success_func=(
                _download_exported_file,
                (client, args.cluster_id, args.save_as, export_job_id),
            ),
        )


def export_jobs_delete(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(
        f"/api/v2/clusters/{args.cluster_id}/export-jobs/{args.export_job_id}/"
    )
    print_response(
        data=data,
        errors=errors,
        keys=["id", "cluster_id", "status"],
        output_fmt=get_output_format(args),
    )


def export_jobs_list(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/clusters/{args.cluster_id}/export-jobs/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "cluster_id", "status", "source", "destination"],
        output_fmt=get_output_format(args),
        transforms={
            "source": _transform_export_job_source,
            "destination": _transform_export_job_destination,
        },
    )


@grand_central_jwt_token
def get_scheduled_jobs(args: Namespace) -> None:
    client = _get_gc_client(args)

    data, errors = client.get("/api/scheduled-jobs/")

    print_response(
        data=data,
        errors=errors,
        keys=["name", "id", "cron", "sql", "enabled", "next_run_time"],
        output_fmt=get_output_format(args),
    )

    if errors or not data:
        return


@grand_central_jwt_token
def get_scheduled_job_log(args: Namespace) -> None:
    client = _get_gc_client(args)

    data, errors = client.get(f"/api/scheduled-jobs/{args.job_id}/log")
    print_response(
        data=data,
        errors=errors,
        keys=["job_id", "start", "end", "error", "statements"],
        output_fmt=get_output_format(args),
    )

    if errors or not data:
        return


@grand_central_jwt_token
def create_scheduled_job(args: Namespace) -> None:
    body = {
        "name": args.name,
        "cron": args.cron,
        "sql": args.sql,
        "enabled": args.enabled,
    }

    client = _get_gc_client(args)

    data, errors = client.post("/api/scheduled-jobs/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["name", "id", "cron", "sql", "enabled"],
        output_fmt=get_output_format(args),
    )

    if errors or not data:
        return


@grand_central_jwt_token
def delete_scheduled_job(args: Namespace) -> None:
    client = _get_gc_client(args)

    data, errors = client.delete(f"/api/scheduled-jobs/{args.job_id}")
    print_response(
        data=data,
        errors=errors,
        success_message="Scheduled job deleted.",
        output_fmt=get_output_format(args),
    )


@grand_central_jwt_token
def edit_scheduled_job(args: Namespace) -> None:
    body = {
        "name": args.name,
        "cron": args.cron,
        "sql": args.sql,
        "enabled": args.enabled,
    }

    client = _get_gc_client(args)

    data, errors = client.put(f"/api/scheduled-jobs/{args.job_id}", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["name", "id", "sql", "cron", "enabled"],
        output_fmt=get_output_format(args),
    )

    if errors or not data:
        return


def _get_gc_client(args: Namespace) -> Client:
    client = Client.from_args(args)
    cluster, _ = client.get(f"/api/v2/clusters/{args.cluster_id}/")

    url_region_cloud = cluster.get("fqdn").split(".", 1)[1][:-1]  # type: ignore
    gc_url = f"https://{cluster.get('name')}.gc.{url_region_cloud}"  # type: ignore
    client.base_url = URL(gc_url)
    client.session.cookies.set("grand_central_session", CONFIG.gc_jwt_token)

    return client


def _transform_export_job_source(field):
    return field["table"]


def _transform_export_job_destination(field):
    return f"Format: {field['format']}\nFile ID: {field.get('file', {}).get('id')}"


# We want to map the custom hardware specs to slightly nicer params in croud,
# hence this mapping here
def _handle_edge_params(body, args):
    if args.cpus:
        body.setdefault("hardware_specs", {})["cpus_per_node"] = args.cpus
    if args.disks:
        body.setdefault("hardware_specs", {})["disks_per_node"] = args.disks
    if args.disk_size_gb:
        body.setdefault("hardware_specs", {})["disk_size_per_node_bytes"] = (
            args.disk_size_gb * 1024 * 1024 * 1024
        )
    if args.disk_type:
        body.setdefault("hardware_specs", {})["disk_type"] = args.disk_type
    if args.memory_size_mb:
        body.setdefault("hardware_specs", {})["memory_per_node_bytes"] = (
            args.memory_size_mb * 1024 * 1024
        )


def _get_operation_status(client: Client, cluster_id: str, request_params: Dict):
    data, errors = client.get(
        f"/api/v2/clusters/{cluster_id}/operations/", params=request_params
    )

    if not data or len(data.get("operations", [])) == 0:
        raise AsyncOperationNotFound("Failed retrieving operation status.")

    operation = data["operations"][0]

    status = operation.get("status")
    feedback_data = operation.get("feedback_data", {})
    msg = feedback_data.get("message", None)

    return status, msg, feedback_data


def _get_import_job_operation_status(
    client: Client, cluster_id: str, request_params: Dict
):
    import_job_id = request_params["import_job_id"]
    data, errors = client.get(
        f"/api/v2/clusters/{cluster_id}/import-jobs/{import_job_id}/"
    )

    if not data or not data.get("progress"):
        raise AsyncOperationNotFound("Failed retrieving operation status.")

    status = data["status"]
    feedback_data = {"progress": data["progress"]}
    msg = data["progress"]["message"]

    return status, msg, feedback_data


def _get_formatted_records_normalized(feedback: dict) -> str:
    num_records = feedback.get("progress", {}).get("records")

    records_normalized = num_records

    if num_records > 1_000_000:
        records_normalized = f"{records_normalized / 1_000_000:.2f}M"
    elif num_records > 1_000:
        records_normalized = f"{records_normalized / 1_000:.2f}K"

    return records_normalized


def _data_job_feedback_func(status: str, feedback: dict, job_type: str):
    records_normalized = _get_formatted_records_normalized(feedback)
    percent = feedback.get("progress", {}).get("percent", 0)
    percent_str = ""
    if percent > 0:
        percent_str = "({:.2f}%) ".format(percent)

    if status == "SUCCEEDED":
        print_info(f"Done {job_type}ing {records_normalized} records")
    else:
        print_info(
            f"{job_type}ing... {records_normalized} records {percent_str}{job_type}ed "
            "so far."
        )


def _download_exported_file(
    client: Client, cluster_id: str, save_as: str, export_job_id: str
):
    data, errors = client.get(
        f"/api/v2/clusters/{cluster_id}/export-jobs/{export_job_id}/"
    )

    if not data or not data.get("progress"):
        raise AsyncOperationNotFound("Failed retrieving operation status.")

    status = data["status"]
    if status == "SUCCEEDED":
        file_id = data.get("destination", {}).get("file", {}).get("id")
        if file_id:
            org_id = _get_org_id_from_cluster_id(client, cluster_id)
            data, errors = client.get(
                f"/api/v2/organizations/{org_id}/files/{file_id}/"
            )
            file_data: dict = cast(dict, data)
            if not (file_data and file_data.get("download_url")):
                print_error("File could not be fetched.")
            if not save_as:
                print_success(f"Download URL: {file_data['download_url']}")
                return
            HALO.stop()
            print_info("Downloading file...")

            r = requests.get(
                file_data["download_url"], stream=True, allow_redirects=True
            )
            if r.status_code != 200:
                r.raise_for_status()
                print_error(
                    f"Request to {file_data['download_url']} returned status code "
                    f"{r.status_code}"
                )
            file_size = int(r.headers.get("Content-Length", 0))

            path = pathlib.Path(save_as).expanduser().resolve()
            path.parent.mkdir(parents=True, exist_ok=True)

            desc = "(Unknown total file size)" if file_size == 0 else ""
            r.raw.read = functools.partial(r.raw.read, decode_content=True)
            with tqdm.wrapattr(r.raw, "read", total=file_size, desc=desc) as r_raw:
                with path.open("wb") as f:
                    copyfileobj(r_raw, f)

            print_success(f"Successfully downloaded file to {path}")


def _get_export_job_operation_status(
    client: Client, cluster_id: str, request_params: Dict
):
    export_job_id = request_params["export_job_id"]
    data, errors = client.get(
        f"/api/v2/clusters/{cluster_id}/export-jobs/{export_job_id}/"
    )

    if not data or not data.get("progress"):
        raise AsyncOperationNotFound("Failed retrieving export operation status.")

    status = data["status"]
    feedback_data = {"progress": data["progress"]}
    msg = data["progress"]["message"]

    return status, msg, feedback_data


def _wait_for_completed_operation(
    *,
    client: Client,
    cluster_id: str,
    request_params: Dict,
    operation_status_func=_get_operation_status,
    feedback_func=None,
    post_success_func=None,
):
    last_status = None
    last_msg = None
    while True:
        try:
            status, msg, feedback = operation_status_func(
                client=client, cluster_id=cluster_id, request_params=request_params
            )
        except AsyncOperationNotFound as e:
            print_error(str(e))
            break

        # Inform about a change in status if the status is not final.
        if status not in ["FAILED", "SUCCEEDED"] and (
            last_status != status or msg != last_msg
        ):
            to_print = f"Status: {status} ({msg})" if msg else f"Status: {status}"
            print_info(to_print)
            last_status = status
            last_msg = msg

        # Call for custom feedback if function available and there is status to report.
        if status in ["IN_PROGRESS", "SUCCEEDED"] and feedback_func:
            (feedback_f, feedback_args) = feedback_func
            feedback_f(status, feedback, *feedback_args)

        # Final statuses
        if status == "SUCCEEDED":
            if post_success_func:
                (func, call_args) = post_success_func
                func(*call_args)
            print_success("Operation completed.")
            break
        if status == "FAILED":
            if msg:
                print_error(msg)
            else:
                print_error(
                    "Your cluster operation has failed. "
                    "Our operations team is investigating the issue."
                )
            break

        with HALO:
            time.sleep(10)


def _lookup_organization_id_for_project(
    client: Client, args: Namespace, project_id: str
) -> Optional[str]:
    data, errors = client.get(f"/api/v2/projects/{project_id}/")
    if not data or errors:
        print_response(
            data=data,
            errors=errors,
            output_fmt=get_output_format(args),
        )
        return None
    return data.get("organization_id")



from unittest import mock

from croud.api import Client, RequestMethod
from tests.util import assert_rest, call_command, gen_uuid


@mock.patch.object(Client, "request", return_value=({}, None))
def test_cloud_configurations_get(mock_request):
    org_id = gen_uuid()
    call_command(
        "croud",
        "cloud-configurations",
        "get",
        "--key",
        "my_config_key",
        "--org-id",
        org_id,
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        "/api/v2/configurations/my_config_key/",
        params={"organization_id": org_id},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_cloud_configurations_set(mock_request):
    org_id = gen_uuid()
    call_command(
        "croud",
        "cloud-configurations",
        "set",
        "--key",
        "my_config_key",
        "--value",
        "new_config_value",
        "--org-id",
        org_id,
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        "/api/v2/configurations/my_config_key/",
        body={"value": "new_config_value", "organization_id": org_id},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_cloud_configurations_list(mock_request):
    org_id = gen_uuid()
    call_command("croud", "cloud-configurations", "list", "--org-id", org_id)
    assert_rest(
        mock_request,
        RequestMethod.GET,
        "/api/v2/configurations/",
        params={"organization_id": org_id},
    )
